#include "filter.h"
#include "../../../common/parser/parser.h"

using namespace MaskTools;
using namespace MaskTools::Picture;
using namespace MaskTools::Parameters;
using namespace MaskTools::Constraints;
using namespace MaskTools::Parser;

namespace MaskTools { namespace Filters { namespace Mask { namespace Edge {

void Filter::ProcessY(PlaneProcessor *processor, Plane &dst, CPlane &src1,
                      CPlane &src2, CPlane &src3)
{
   static_cast<Processor*>(processor)(dst.plane(), dst.pitch(), src1.plane(), src1.pitch(), matrix, nLowThresholdY, nHighThresholdY, nWidth, nHeight);
}

void Filter::ProcessU(PlaneProcessor *processor, Plane &dst, CPlane &src1,
                      CPlane &src2, CPlane &src3)
{
   static_cast<Processor*>(processor)(dst.plane(), dst.pitch(), src1.plane(), src1.pitch(), matrix, nLowThresholdUV, nHighThresholdUV, nWidthUV, nHeightUV);
}

void Filter::ProcessV(PlaneProcessor *processor, Plane &dst, CPlane &src1,
                      CPlane &src2, CPlane &src3)
{
   static_cast<Processor*>(processor)(dst.plane(), dst.pitch(), src1.plane(), src1.pitch(), matrix, nLowThresholdUV, nHighThresholdUV, nWidthUV, nHeightUV);
}


Filter::Filter(const Parameters::Parameters &parameters) : MaskTools::Filter(parameters)
{
   nLowThresholdY = clip(parameters["thY1"].toInt(), 0, 255);
   nHighThresholdY = clip(parameters["thY2"].toInt(), 0, 255);
   nLowThresholdUV = clip(parameters["thC1"].toInt(), 0, 255);
   nHighThresholdUV = clip(parameters["thC2"].toInt(), 0, 255);

   /* add the processors */
   if ( parameters["mode"].toString() == "sobel" )
   {
      print(LOG_INFO, "Edge : using sobel detector");
      addProcessor(Constraints::Processor<PlaneProcessor>(PPlaneProcessor(sobel_c), Constraint(CPU_NONE, 1, 1, 1, 1), 0));
      addProcessor(Constraints::Processor<PlaneProcessor>(PPlaneProcessor(sobel8_mmx), Constraint(CPU_MMX, 8, 1, 1, 1), 1));
   }
   else if ( parameters["mode"].toString() == "roberts" )
   {
      print(LOG_INFO, "Edge : using roberts detector");
      addProcessor(Constraints::Processor<PlaneProcessor>(PPlaneProcessor(roberts_c), Constraint(CPU_NONE, 1, 1, 1, 1), 0));
      addProcessor(Constraints::Processor<PlaneProcessor>(PPlaneProcessor(roberts8_mmx), Constraint(CPU_MMX, 8, 1, 1, 1), 1));
   }
   else if ( parameters["mode"].toString() == "laplace" )
   {
      print(LOG_INFO, "Edge : using laplace detector");
      addProcessor(Constraints::Processor<PlaneProcessor>(PPlaneProcessor(laplace_c), Constraint(CPU_NONE, 1, 1, 1, 1), 0));
      addProcessor(Constraints::Processor<PlaneProcessor>(PPlaneProcessor(laplace8_mmx), Constraint(CPU_MMX, 8, 1, 1, 1), 1));
   }
   else
   {
      print(LOG_INFO, "Edge : using custom detector");
      std::list<Symbol> coeffs = Parser::getDefaultParser().parse(parameters["mode"].toString(), " ").getExpression();
      memset(matrix, 0, sizeof(matrix));
      for ( int i = 0; i < 9; i++ )
      {
         if ( !coeffs.size() )
         {
            error = "Edge : unvalid custom detector";
            return;
         }
         matrix[9] += abs(matrix[i] = Word(coeffs.front().getValue(0,0,0)));
         coeffs.pop_front();
      }

      if ( coeffs.size() )
         matrix[9] = Word(coeffs.front().getValue(0,0,0));
      else
      {
         /* choose the upper power of 2, if possible */
         for ( int i = 0; i < 8; i++ )
            if ( (1 << i) > matrix[9] )
            {
               matrix[9] = 1 << i;
               break;
            }
      }

      if ( !matrix[9] )
         matrix[9] = 1;

      print(LOG_INFO, "Edge : %3i %3i %3i\n"
                      "       %3i %3i %3i\n"
                      "       %3i %3i %3i\n"
                      "normalize: %3i", matrix[0], matrix[1], matrix[2], matrix[3], matrix[4], matrix[5], matrix[6], matrix[7], matrix[8], matrix[9]);
      addProcessor(Constraints::Processor<PlaneProcessor>(PPlaneProcessor(convolution_c), Constraint(CPU_NONE, 1, 1, 1, 1), 0));
      /* insert mmx version only if matrix[9] is a valid power of 2 */
      if ( matrix[9] == 1 || matrix[9] == 2 || matrix[9] == 4 || matrix[9] == 8 || matrix[9] == 16 || matrix[9] == 32 || matrix[2] == 64 )
         addProcessor(Constraints::Processor<PlaneProcessor>(PPlaneProcessor(convolution8_mmx), Constraint(CPU_MMX, 8, 1, 1, 1), 1));
   }
}

String Filter::getName() { return "mt_edge"; }

Signature Filter::getSignature()
{
   Signature signature;
   signature.push_back(Parameter(CLIP, ""));
   signature.push_back(Parameter(String("sobel"), "mode"));
   signature.push_back(Parameter(10, "thY1"));
   signature.push_back(Parameter(10, "thY2"));
   signature.push_back(Parameter(10, "thC1"));
   signature.push_back(Parameter(10, "thC2"));

   return addDefaultCommons(signature);
}

Clip::Inputs Filter::inputClips = Clip::OneFrame();

} } } }
